// License: GPL. For details, see Readme.txt file.
package org.openstreetmap.gui.jmapviewer.interfaces;

public interface ICoordinate {

    double getLat();

    void setLat(double lat);

    double getLon();

    void setLon(double lon);
}
